#!/usr/bin/perl -w


#
#  Copyright (c) 2004-2005 The University of Chicago, as Operator
#  of Argonne National Laboratory.
#

# network_crawler.pl
#   Reads  accel_sw.log files,and writes resulting network data to db.
#
# Authors: Janet Anderson (jba@aps.anl.gov)
#
#
#

use Mail::Mailer;
use File::stat;
use Cwd;
use Blowfish_PP;

use LogUtil;
use NetworkCrawlerParser;
use NetworkCrawlerDBLayer;

my $status = 0;
my $allowed_updates = 4;

##########################################################
#  GET SCRIPT ARGUMENTS
##########################################################
my $test_mode = 0;
my $go = 0; # don't run script unless explicitly enabled
my $debug_level = "info"; # default
my $decryption_key = "";
my $replace_all_network_data = 0;
my $force_database_update = 0; # don't update database unless than 4 updates

if (!@ARGV) {
    print_usage();
    exit;
}
while (@ARGV) {
    $command = shift @ARGV;
    if ($command =~ /-help/) {
        print_usage();
        exit;
    } elsif  ($command =~ /--force/) {
        $force_database_update = 1;
    } elsif  ($command =~ /--all/) {
        $replace_all_network_data = 1;
    } elsif ($command =~ /--debug-level=(.*)/) {
        $debug_level = $1;
    } elsif ($command =~ /--go/) {
        $go = 1;
    } elsif ($command =~ /--key=(.*)/) {
        # Optional 16 byte blowfish decryption key. If given (non null),
        $decryption_key = $1;
    } elsif ($command =~ /--test/) {
        set_test_mode(1);
        $test_mode =1;
    } else {
        print_usage();
        exit;
    }
}
if (!$go ) {
    print "You must at a minimum use --go command line option to run the script!\n";
    exit;
}

##########
#  SETUP
##########

# Can be 'verbose', 'debug', 'info', 'warn', or 'error'
LogUtil::set_log_level($debug_level);

email_log('info',"Starting ioc network data update script.");   

# Set the input and temporary files
$infile            = "/usr/local/iocapps/iocinfo/network/accel_sw.log";
$dir               = "/tmp";
# =======  Debug settings
#$infile            = "./accel_sw.log";
#$dir               = "/home/phoebus/JBA/irmis/irmisBase/db/src/crawlers/network";
# =======
my $prev_infile       = "./accel_sw.log_save";
my $infileDiff        = "$dir/accel_sw.log_diff";
$infileDiffSorted  = "$dir/accel_sw.log_diff_sorted";

$sb_in=stat($infile);
$infile_date=scalar localtime $sb_in->mtime;

if ( ! -f $prev_infile ) {
  email_log('info',"$prev_infile not found. Replacing ALL network data.");   
  $replace_all_network_data = 1;
}

################
#  Setup DBLayer
################

# Tell db layer where the home dir is before we do any chdir.
set_home_dir(getcwd());

# it is assumed the db.properties file is encrypted.
set_decryption_key($decryption_key);


#######################################################
#  Start database update or replace of ioc network data
#######################################################

my @network_hits=();

if ($replace_all_network_data) {

  #########################################################################
  # Replace all ioc network data with data from latest accel_sw.log file
  #########################################################################
  email_log('info',"New accel_sw.log file date: $infile_date");   
  $status = zero_out_network_data();
  if ( $status ==  0 ) {
    $status = get_all_network_data($infile,\@network_hits);
    if ( $status ==  0 ) {
      $status = load_new_network_data(\@network_hits);
    }
  }

} else {

  #############################################################################
  # Update ioc network data from accel_sw.log if differs from prev_accel_sw.log
  #############################################################################

  email_log('info',"Updating modified network data.");   

  $sb_prev_in=stat($prev_infile);
  $prev_infile_date=scalar localtime $sb_prev_in->mtime;

  $status = create_diff_file();
  if ( $status == 1) {
    email_log('info',"No data updated.");   
  }
  if ( $status == 0) {
    $status = get_modified_network_data($infileDiffSorted,\@network_hits);
    if ( $status >=  0 ) {
      if ( $status > $allowed_updates && !$force_database_update ) {
        email_log('error',"Update count $status is more than $allowed_updates. Database NOT updated.");   
      } else { 
        $status = load_new_network_data(\@network_hits);
      }
    }
  }
  unlink $infileDiff if -f $infileDiff ;
  #unlink $infileDiffSorted if -f $infileDiffSorted ;
}

#########################
#  Save network data file
#########################

if ( !$test_mode ) {
  if ( $status == 0 ) {
    unlink $prev_infile if -f $prev_infile;
    $status=system("cp -p $infile $prev_infile");
    if ($status != 0) {
      email_log('error',"Copy of $infile failed. $?");   
    } else {
      email_log('info',"Saved copy of $infile.");   
    }
  } else {
    email_log('warn',"$infile NOT copied.");
  }
}

##################################
# Finish log, send email and exit.
##################################

email_log('info',"Finished ioc network data update script.");   
send_email();

exit;

#########################################################################
# Create sorted file of differences
#########################################################################
sub create_diff_file {

  if ( ! -f $infile ) {
    email_log('error',"$infile not found.");   
    return -1;
  }
  email_log('info',"New accel_sw.log file date: $infile_date");   
  email_log('info',"Saved accel_sw.log file date: $prev_infile_date");   

  unlink $infileDiff if -f $infileDiff;
  system("diff $prev_infile $infile > $infileDiff");
  if ( !-e $infileDiff ) { 
    email_log('error',"Network data differences file not created.");
    return -1;
  }
  if ( -z  $infileDiff ) {
    email_log('info',"No network data differences found.");
    return 1;
  }
  unlink $infileDiffSorted;
  if ( system("sort -k 4.1b,4 -k 1,1  $infileDiff > $infileDiffSorted") != 0 ) {
    email_log('error',"Cannot sort differences file $infileDiff");
    return -1;
  }
  return 0;
}

#########################################################################
# Print out application usage text
#########################################################################
sub print_usage {
    print "\n";
    print "usage: network_crawler --go [--key=<16 char key>] [--test/all/force] [--help] [--debug-level=[verbose|debug|info]]\n\n";
    print "The network_crawler reads a file of network switch interface data parsing\n";
    print "the data and comparing it to data in a saved file from a previous run to\n";
    print "determine if there are changes to the ioc primary and secondary ethernet\n";
    print "switch,blade, and port settings. If there are less than $allowed_updates changes\n";
    print "the changes are then inserted into the mysql database using perl DBI.\n";

    print "Log output goes to standard out, and can be adjusted from the default\n";
    print "\"info\" level. Watch out becuase verbose is VERY verbose.\n\n";

    print "You will need to establish the expected schema by running the\n";
    print "create_pv_tables.sql and alter_pv_tables.sql scripts in the rdbCore\n";
    print "ddl subdirectory. Then edit db.properties to set your local database\n";
    print "connection parameters (ie. user/password etc.).\n\n";

    print "The --test option runs the crawler, but does no db changes.\n";
    print "\n";
    print "The --all option replaces all ioc network data with data from the\n";
    print "latest accel_sw.log file.\n";
    print "\n";
    print "The --force option forces a database update when there are more\n";
    print "than the allowed number of database updates ($allowed_updates).\n";
    print "\n";

    print "You must have --go command line option to run the script!\n\n";
}

